#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <time.h>
#include <linux/i2c-dev.h>
#include "ice40-pins.h"
#include "eim.h"

static uint16_t *ice40_out_reg;
static uint16_t *ice40_in_reg;
#define ICE40_WRITE(x) (*ice40_out_reg = (x))
#define ICE40_READ (*ice40_in_reg)
//define SHORT_SLEEP nanosleep((const struct timespec[]){0, 50}, NULL)
#define SHORT_SLEEP ;

/* Code to clock out a single byte. */
void send_byte(unsigned char b) {
  char bit, i;
  for (i = 0; i < 8; i++) {
    bit = b & 128 ? ICE40_SDI : 0;
    ICE40_WRITE( ICE40_SS|ICE40_RESET|bit);
    SHORT_SLEEP;
    ICE40_WRITE( ICE40_SS|ICE40_SCK|ICE40_RESET|bit);
    SHORT_SLEEP;
    b = (b & 0x7F) << 1;
  }
}

void *load_bitfile(char *filename, int *length) {
  int bitfile_fd;
  int result;
  int file_size;
  void *bitfile_buffer;
  int fpos;

  bitfile_fd = open(filename, O_RDONLY);
  if (bitfile_fd < 0) {
    fprintf(stderr, "Failed to open bit file.\n");
    perror(filename);
    exit(2);
  }

  /* Measure the length of the file and allocate a suitable buffer. */
  result = lseek(bitfile_fd, 0L, SEEK_END);
  if (result < 0 ) {
   fprintf(stderr, "Failed to seek bit file.\n");
    perror(filename);
    exit(2);
  }
  
  file_size = result;
  result = lseek(bitfile_fd, 0L, SEEK_SET);
  if (result < 0 ) {
    fprintf(stderr, "Failed to seek bit file.\n");
    perror(filename);
    exit(2);
  }

  /* Allocate a buffer and read in the file. */
  bitfile_buffer = malloc(file_size);
  if (bitfile_buffer == NULL) {
    fprintf(stderr, "Failed to allocate memory for bit file.\n");
    perror(filename);
    exit(2);
  }

  fpos = 0;
  while (fpos < file_size) {
    result = read(bitfile_fd, bitfile_buffer+fpos, file_size-fpos);
    if (result < 0) {
      fprintf(stderr, "Failed to read from bit file.\n");
      perror(filename);
      exit(2);
    }
    fpos += result;
  }
  
  close(bitfile_fd);

  *length = file_size;
  return bitfile_buffer;
}

int main(int argc, char **argv) {
  int i, file_size;
  void *bitfile_buffer;
  if (argc != 2) {
    fprintf(stderr, "Syntax: %s <bit file>\n", argv[0]);
    return 1;
  }

  /* Load the whole file into memory, because
   * the iCE40 says we're not allowed to stall
   * while feeding in the data. This appears
   * to be a lie, but let's humour it. */
  bitfile_buffer = load_bitfile(argv[1], &file_size);

  /* Initialise EIM. */
  ice40_in_reg = eim_get(fpga_r_ice40_in);
  ice40_out_reg = eim_get(fpga_w_ice40_out);
    
  /* Put the iCE40 into reset, and set the ICE40_SS pin low so that it comes up
   * in SPI slave mode, ICE40_SCK high because that's what the manual says. :) */
  ICE40_WRITE( ICE40_SCK);
  
  /* Sleep 200us to let the iCE40 notice the reset. */
  nanosleep((const struct timespec[]){0, 200000}, NULL);

  /* Bring the chip up out of reset. */
  ICE40_WRITE( ICE40_SCK|ICE40_RESET);

  /* Sleep 1500us to let the iCE40 reset itself.. */
  nanosleep((const struct timespec[]){0, 1500000}, NULL);

  /* Let the slave select go high again, and send out 8 clocks.
   * Nominal clock frequency is 10MHz but that's not going to happen
   * over I2C, so don't take the sleep numbers seriously. */
  ICE40_WRITE( ICE40_SS|ICE40_SCK|ICE40_RESET);

  /* Send 8 clocks, because apparently it needs 8 clocks... */
  for (i = 0; i < 8; i++) {
    ICE40_WRITE( ICE40_SS|ICE40_RESET);
    SHORT_SLEEP;
    ICE40_WRITE( ICE40_SS|ICE40_SCK|ICE40_RESET);
    SHORT_SLEEP;
  }

  /* Clock out the actual data from the file. */
  for (i = 0; i < file_size; i++) {
    send_byte(((char *)bitfile_buffer)[i]);
  }

  /* Send another 100 clocks to finish starting up the FPGA. */
  for (i = 0; i < 100; i++) {
    ICE40_WRITE( ICE40_SS|ICE40_RESET);
    SHORT_SLEEP;
    ICE40_WRITE( ICE40_SS|ICE40_SCK|ICE40_RESET);
    SHORT_SLEEP;
  }

  /* Check ICE40_CDONE. */
  if (ICE40_READ & ICE40_CDONE) {
    printf("Success.\n");
  }
  else {
    fprintf(stderr,"Failed to start iCE40\n");
    for (i=0; i<100; i++) {
      fprintf(stderr, "%x ", ICE40_READ);
    }
    return 5;
  }
  free(bitfile_buffer);
}
